#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "twi/twi.h"
#include "usbconfig.h"
#include "usbdrv/usbdrv.h"

// Wii GamePad slave address
#define WIIEXT_TWI_ADDR 0x52

// for send data to Macintosh/PC
static struct gamepad_report_t
{
	uint16_t buttons;
	int8_t left_x;
	int8_t left_y;
	int8_t right_x;
	int8_t right_y;
} gamepad_report;
//I2C TX/RX Buffer
uint8_t twiBuffer[8];
//KeyData&XY Data & Flag
unsigned char status[6];
unsigned char buttons[2];
unsigned char  AxisInit;
unsigned char  nunchak_flag;
//XY Axis
int centerLeftX,rawLeftX;
int centerLeftY,rawLeftY;
int centerRightX,rawRightX;
int centerRightY,rawRightY;
// loop counter
int cnt ;

const PROGMEM char usbHidReportDescriptor[46] = {
// make sure the size matches USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH in usbconfig.h
	0x05, 0x01,	// USAGE_PAGE (Generic Desktop)
	0x09, 0x05,	// USAGE (Game Pad)
	0xa1, 0x01,	// COLLECTION (Application)
	0xa1, 0x00,	//   COLLECTION (Physical)
	0x05, 0x09,	//	 USAGE_PAGE (Button)
	0x19, 0x01,	//	 USAGE_MINIMUM (Button 1)
	0x29, 0x10,	//	 USAGE_MAXIMUM (Button 16)
	0x15, 0x00,	//	 LOGICAL_MINIMUM (0)
	0x25, 0x01,	//	 LOGICAL_MAXIMUM (1) 
	0x95, 0x10,	//	 REPORT_COUNT (16)
	0x75, 0x01,	//	 REPORT_SIZE (1)
	0x81, 0x02,	//	 INPUT (Data,Var,Abs)
	0x05, 0x01,	//	 USAGE_PAGE (Generic Desktop)
	0x09, 0x30,	//	 USAGE (X) // left X
	0x09, 0x31,	//	 USAGE (Y) // left Y
	0x09, 0x32,	//	 USAGE (Z) // right X
	0x09, 0x33,	//	 USAGE (Rx) // right Y
	0x15, 0x81,	//	 LOGICAL_MINIMUM (-127)
	0x25, 0x7f,	//	 LOGICAL_MAXIMUM (127)
	0x75, 0x08,	//	 REPORT_SIZE (8)
	0x95, 0x04,	//	 REPORT_COUNT (4)
	0x81, 0x02,	//	 INPUT (Data,Var,Abs)
	0xc0,		//   END_COLLECTION
	0xc0		// END_COLLECTION
};

// not used for game pads
static unsigned char idleRate;

unsigned char usbFunctionSetup(unsigned char data[8])
{
	usbRequest_t *rq = (void *)data;

	// there are several request types but we only need to handle
	// 3 cases of the "class" type
	
	if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) /* class request type */
	{
		if (rq->bRequest == USBRQ_HID_GET_REPORT) /* wValue: ReportType (highbyte), ReportID (lowbyte) */
		{			
			usbMsgPtr = &gamepad_report;
			return sizeof(gamepad_report);
		}
		else if (rq->bRequest == USBRQ_HID_GET_IDLE)
		{
			usbMsgPtr = &idleRate;
			return 1;
		}
		else if (rq->bRequest == USBRQ_HID_SET_IDLE)
		{
			idleRate = rq->wValue.bytes[1];
		}
	}
	else
	{
		// no vendor specific requests implemented
	}
	
	return 0;
}

//decryption Wii data
static char nunchuk_decode_byte (char x)
{
    x = (x ^ 0x17) + 0x17;
    return x;
}

// Caribrate XY Axis StartUp
void AxisInitDefoult(void)
{
	centerLeftX = rawLeftX;
	centerLeftY = rawLeftY;
	centerRightX = rawRightX;
	centerRightY = rawRightY;
}

int main()
{	
	DDRC &= ~(_BV(5) | _BV(4)); // TWI pins as input
	PORTC |= _BV(5) | _BV(4); // enable pull-ups
	
	twi_init();
	sei();
	

	
	nunchak_flag=2;
	while(nunchak_flag==2)
	{
	_delay_ms(25); // power up delay
	
	// initialize the Wii Classic Controller
	twiBuffer[0] = 0x40; twiBuffer[1] = 0x00; twiBuffer[2] = 0x00; 
	twi_writeTo(WIIEXT_TWI_ADDR, twiBuffer, 2, 1);
	_delay_us(500);
	
	// Detect Wii Classic Controller
	twiBuffer[0] = 0xFE;
	twi_writeTo(WIIEXT_TWI_ADDR, twiBuffer, 1, 1);
	_delay_ms(15);
	twi_readFrom(WIIEXT_TWI_ADDR, twiBuffer, 2);
	if ((twiBuffer[0] == 0xFD)&&(twiBuffer[1] == 0xFD)) {
		nunchak_flag=0; // find classiccontroller !!
	}
	if ((twiBuffer[0] == 0xFE)&&(twiBuffer[1] == 0xFE)) {
		nunchak_flag=1;// find Nunchak !!
	}

	}
	

	cli();
	
	usbInit(); // start v-usb
	usbDeviceDisconnect(); // enforce USB re-enumeration, do this while interrupts are disabled!
	_delay_ms(250);
	usbDeviceConnect();
	
	sei(); // enable interrupts
	AxisInit=0;//XYstick calibration flag

// First read command
		twiBuffer[0] = 0x00;
		twi_writeTo(WIIEXT_TWI_ADDR, twiBuffer, 1, 1);
		_delay_us(300);

	while (1) // forever loop
	{
		
		usbPoll();
		
		// read raw values
		twi_readFrom(WIIEXT_TWI_ADDR, twiBuffer, 6);
		//Decription data
		for (cnt = 0; cnt <= 5 ; cnt++) {
            if (cnt < 4) {
                status[cnt] = nunchuk_decode_byte (twiBuffer[cnt]);
            } else {
            	status[cnt] = nunchuk_decode_byte (twiBuffer[cnt]);
                buttons[cnt-4] = nunchuk_decode_byte (twiBuffer[cnt]);
            }
    	}
		if(nunchak_flag==1)//Nunchak
		{
			rawLeftX = (int)status[0] ;
			rawLeftY = (int)status[1] ;
		}
		else							//Classic Controller
		{
			rawLeftX = (int)((status[0] & 0x3F) << 2);
			rawLeftY = (int)((status[1] & 0x3F) << 2);
		}
		rawRightX = (int)((status[0] & 0xC0) | ((status[1] & 0xC0) >> 2) | ((status[2] & 0x80) >> 4));
		rawRightY = (int)((status[2] & 0x1F) << 3);
		
		//caribrate first recieve data XY axis
		if(AxisInit==0)
		{
			AxisInitDefoult();
			AxisInit=1;
		}
		
		// calculate offset from original center position
		// set the values into the report
		gamepad_report.left_x = rawLeftX - centerLeftX;
		gamepad_report.right_x = rawRightX - centerRightX;
		gamepad_report.left_y = rawLeftY - centerLeftY;
		gamepad_report.right_y = rawRightY - centerRightY;
		gamepad_report.buttons = ~((buttons[0]<< 8) | buttons[1]);
		
		//mask Nunchak unused Key
		if(nunchak_flag==1)
		{
			gamepad_report.buttons &=0x0003;
			gamepad_report.right_x=gamepad_report.right_y=0x0000;
		}
		
		// Read command
		twiBuffer[0] = 0x00;
		twi_writeTo(WIIEXT_TWI_ADDR, twiBuffer, 1, 1);
		
		// wait until endpoint is ready
		while (!usbInterruptIsReady()) {
			usbPoll();
		}
		
		// send report
		usbSetInterrupt((unsigned char *)(&gamepad_report), sizeof(gamepad_report));
	}
	
	return 0;
}
